"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ClinicalTrialsGovAPI = exports.isPagedStudies = exports.DEFAULT_ENDPOINT = void 0;
/**
 * Module for invoking the Clinical Trials.gov API.
 */
const util_1 = require("util");
const node_fetch_1 = require("node-fetch");
/**
 * The default endpoint if none is specified: the clinicaltrials.gov v2 API
 * endpoint.
 */
exports.DEFAULT_ENDPOINT = 'https://clinicaltrials.gov/api/v2';
/**
 * Type guard for checking the response from the server. Exported for tests.
 * @param o the object to check
 * @returns
 */
function isPagedStudies(o) {
    if (typeof o !== 'object' || o === null) {
        return false;
    }
    const maybeStudies = o.studies;
    if (!Array.isArray(maybeStudies)) {
        return false;
    }
    // Make sure the studies array are all at least valid enough to not crash everything
    return maybeStudies.every((e) => typeof e === 'object' && e !== null);
}
exports.isPagedStudies = isPagedStudies;
/**
 * Class that provides a method for invoking the ClinicalTrials.gov V2 API using
 * the built-in Node.js HTTPS library.
 */
class ClinicalTrialsGovAPI {
    constructor(options) {
        var _a, _b;
        /**
         * Maximum number of redirects before giving up.
         */
        this._maxRedirects = 10;
        this._log =
            (_a = options === null || options === void 0 ? void 0 : options.logger) !== null && _a !== void 0 ? _a : (0, util_1.debuglog)('ctgov-api', /* istanbul ignore next: not worth the effort to mock/test */ (log) => {
                this._log = log;
            });
        this._endpoint = (_b = options === null || options === void 0 ? void 0 : options.endpoint) !== null && _b !== void 0 ? _b : exports.DEFAULT_ENDPOINT;
    }
    /**
     * Fetches the given list of NCT IDs.
     * @param nctIds the NCT IDs to fetch
     * @returns a promise that resolves to the list of studies
     */
    fetchStudies(nctIds, pageSize) {
        return __awaiter(this, void 0, void 0, function* () {
            let url = `${this._endpoint}/studies?filter.ids=${nctIds.join(',')}`;
            if (pageSize && pageSize > 0) {
                url += `&pageSize=${pageSize}`;
            }
            const studies = [];
            let nextPageToken;
            do {
                const result = yield this.getUrl(nextPageToken ? `${url}&pageToken=${nextPageToken}` : url);
                if (isPagedStudies(result)) {
                    studies.push(...result.studies);
                    nextPageToken = result.nextPageToken;
                }
                else {
                    throw new Error('Server returned a success response, but the result could not be parsed.');
                }
            } while (nextPageToken);
            return studies;
        });
    }
    getUrl(url) {
        return __awaiter(this, void 0, void 0, function* () {
            this._log('Fetching [%s]', url);
            const response = yield (0, node_fetch_1.default)(url);
            if (response.ok) {
                return response.json();
            }
            else {
                throw new Error(`Error from ClinicalTrials.gov server: ${response.status} ${response.statusText}`);
            }
        });
    }
}
exports.ClinicalTrialsGovAPI = ClinicalTrialsGovAPI;
exports.default = ClinicalTrialsGovAPI;
