"use strict";
/**
 * This file contains a backup system for finding necessary trial information if
 * your matching service does not provide it.
 *
 * The intended usages is essentially:
 *
 * Create the service:
 *
 * const ctgService = await createClinicalTrialsGovService('temp-data');
 *
 * Then, when ResearchStudy objects exist that need to be filled out, assuming
 * that studies is a ResearchStudy[]:
 *
 * await ctgService.updateResearchStudies(studies);
 *
 * This will fill out whatever can be filled out within the given studies.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createClinicalTrialsGovService = exports.ClinicalTrialsGovService = exports.mkdir = exports.CacheEntry = exports.parseStudyJson = exports.findNCTNumbers = exports.findNCTNumber = exports.isValidNCTNumber = exports.CLINICAL_TRIAL_IDENTIFIER_CODING_SYSTEM_URL = void 0;
const fs = require("node:fs");
const path = require("node:path");
const util_1 = require("util");
const clinicaltrialsgov_api_1 = require("./clinicaltrialsgov-api");
const study_fhir_converter_1 = require("./study-fhir-converter");
/**
 * The URL for the "system" field of identifiers that indicate the
 * ClinicalTrial.gov identifier, or NTC number, of a clinical trial.
 */
exports.CLINICAL_TRIAL_IDENTIFIER_CODING_SYSTEM_URL = 'http://clinicaltrials.gov/';
/**
 * Checks to ensure that a given NCT number is valid. This checks to make sure
 * it matches the regular expression /^NCT[0-9]{8}$/.
 * @param nctNumber the NCT number to check
 */
function isValidNCTNumber(nctNumber) {
    return /^NCT[0-9]{8}$/.test(nctNumber);
}
exports.isValidNCTNumber = isValidNCTNumber;
/**
 * Finds the NCT number specified for the given ResearchStudy, assuming there is
 * one. This requires an identifier on the given ResearchStudy that either
 * belongs to the coding system "http://clinicaltrials.gov/" or is a valid NCT
 * number as determined by #isValidNCTNumber.
 *
 * This will return the value of the first identifier belonging to the coding
 * system "http://clinicaltrials.gov/", and if there is no identifier, return
 * the first identifier value that is considered valid.
 *
 * @param study the research study
 * @returns the NCT number or null if none
 */
function findNCTNumber(study) {
    if (study.identifier && Array.isArray(study.identifier) && study.identifier.length > 0) {
        for (const identifier of study.identifier) {
            if (identifier.system === exports.CLINICAL_TRIAL_IDENTIFIER_CODING_SYSTEM_URL && typeof identifier.value === 'string')
                return identifier.value;
        }
        // Fallback: regexp match
        for (const identifier of study.identifier) {
            if (typeof identifier.value === 'string' && isValidNCTNumber(identifier.value))
                return identifier.value;
        }
    }
    // Return null on failures
    return null;
}
exports.findNCTNumber = findNCTNumber;
/**
 * Finds all the NCT numbers within the given list of studies. Returns a map of
 * NCT numbers to the research studies they match. If multiple research studies
 * have the same NCT ID, then an array will be used to contain all matching
 * studies. (This should be very uncommon but is supported anyway.)
 * @param studies the NCT numbers found, if any
 */
function findNCTNumbers(studies) {
    const result = new Map();
    for (const study of studies) {
        const nctId = findNCTNumber(study);
        if (nctId !== null) {
            const existing = result.get(nctId);
            if (existing === undefined) {
                result.set(nctId, study);
            }
            else {
                if (Array.isArray(existing)) {
                    existing.push(study);
                }
                else {
                    result.set(nctId, [existing, study]);
                }
            }
        }
    }
    return result;
}
exports.findNCTNumbers = findNCTNumbers;
/**
 * Attempts to parse the given string as a Study JSON object.
 * @param fileContents the contents of the file as a string
 * @param log an optional logger
 * @returns a promise that resolves to the parsed object or null if the study
 * could not be parsed
 */
function parseStudyJson(fileContents, log) {
    try {
        const json = JSON.parse(fileContents);
        if (typeof json === 'object' && !Array.isArray(json)) {
            // "null" is currently "valid" - a cached failure?
            return json;
        }
        else {
            if (log) {
                log('Invalid JSON object for Study: %o', json);
            }
            return null;
        }
    }
    catch (ex) {
        if (log) {
            log('Unable to parse JSON object: %o', ex);
        }
        return null;
    }
}
exports.parseStudyJson = parseStudyJson;
/**
 * A cache entry. Cache entries basically operate in two modes: an entry that is pending being written based on a ZIP
 * file, and a file that has a backing file ready to be loaded.
 *
 * The pending state is kind of weird because an entry is pending once a download has been requested for it, but cannot
 * be properly fulfilled until the file data has been saved. There's no really good way to wrap that in a Promise so
 * instead the resolve method is stored until the pending state can be resolved.
 */
class CacheEntry {
    /**
     * Create a new cache entry.
     * @param filename the file name of the entry
     * @param options the file stats (if the file exists) or a flag indicating it's pending (being downloaded still)
     */
    constructor(cache, filename, options) {
        this.filename = filename;
        /**
         * The pending status. Either a Promise (in which case it is not only pending, but has things actively waiting for
         * it), or true, in which case it is pending but the promise is yet to be created, or false if pending but the promise
         * has not been created. This is so that no Promise is created if nothing ever needs it, which avoids cases where the
         * Promise won't have a catch handler attached to it.
         */
        this._pending = false;
        this._cache = cache;
        if (options.stats) {
            // Default to using the metadata from the fs.Stats object
            this._createdAt = options.stats.ctime;
            // Assume the last modified time was when the cache entry was fetched (it's close enough anyway)
            this._lastAccess = options.stats.mtime;
        }
        else {
            // Otherwise, default to now
            this._createdAt = new Date();
            this._lastAccess = new Date();
        }
        if (options.pending) {
            // In this mode, mark createdAt as null
            this._createdAt = null;
            this._pending = true;
        }
    }
    // As the returned Date objects are mutable, return copies
    /**
     * Gets the time when this entry was initially created. If null, that indicates that the entry was never successfully
     * created - it's still waiting for data to be downloaded.
     */
    get createdAt() {
        return this._createdAt === null ? null : new Date(this._createdAt);
    }
    /**
     * Gets the last time the cache entry was accessed. This is updated whenever
     * #load() is called.
     */
    get lastAccess() {
        return new Date(this._lastAccess);
    }
    /**
     * Determine if the cache entry is still pending: data for it has not yet been saved.
     */
    get pending() {
        return this._pending !== false;
    }
    /**
     * Check if the last access time is before the given date
     * @param date the date to check
     * @returns true if the last access time was before the given date
     */
    lastAccessedBefore(date) {
        return this._lastAccess < date;
    }
    /**
     * Indicates that the entry has been located somewhere and that data for it is now being prepared.
     */
    found() {
        if (this._createdAt === null) {
            this._createdAt = new Date();
        }
    }
    /**
     * Resolves the pending state (if the entry was pending), otherwise does nothing. This does not change the "found"
     * status - if ready() is called without found(), createdAt remains null and the entry may be removed as pointing to
     * a record that does not exist.
     */
    ready() {
        if (typeof this._pending === 'object') {
            this._pending.resolve();
        }
        this._pending = false;
    }
    /**
     * Forcibly fail the entry.
     * @param e the error to fail the entry with
     */
    fail(e) {
        if (typeof this._pending === 'object') {
            if (typeof e === 'string') {
                e = new Error(e);
            }
            this._pending.reject(e);
        }
        this._pending = false;
    }
    /**
     * Loads the underlying file. If the entry is still pending, then the file is read once the entry is ready. This may
     * resolve to null if the clinical study does not exist in the results.
     */
    load() {
        // Move last access to now
        this._lastAccess = new Date();
        // If we're still pending, we have to wait for that to finish before we can
        // read the underlying file.
        if (this._pending) {
            let promise;
            if (this._pending === true) {
                // This is the only case when we actually create the Promise - when still pending and something attempts a load.
                // TODO (maybe): Add a timeout?
                // There's no way to prove it to TypeScript, but the executor function in the Promise runs immediately when the
                // Promise is created. So create useless no-op funcs to avoid "may be undefined" errors.
                let resolveFunc = /* istanbul ignore next */ () => {
                    /* no-op */
                }, rejectFunc = /* istanbul ignore next */ () => {
                    /* no-op */
                };
                promise = new Promise((resolve, reject) => {
                    resolveFunc = resolve;
                    rejectFunc = reject;
                });
                this._pending = {
                    promise: promise,
                    resolve: resolveFunc,
                    reject: rejectFunc
                };
            }
            else {
                promise = this._pending.promise;
            }
            return promise.then(() => {
                return this.readFile();
            });
        }
        else {
            // Otherwise we can just return immediately
            return this.readFile();
        }
    }
    /**
     * Always attempt to read the file, regardless of whether or not the entry is pending.
     */
    readFile() {
        return new Promise((resolve, reject) => {
            this._cache.fs.readFile(this.filename, { encoding: 'utf8' }, (err, data) => {
                if (err) {
                    reject(err);
                }
                else {
                    // Again bump last access to now since the access has completed
                    this._lastAccess = new Date();
                    if (data.length === 0) {
                        // Sometimes files end up empty - this appears to be a bug?
                        // It's unclear what causes this to happen
                        this._cache.log('Warning: %s is empty on read', this.filename);
                        resolve(null);
                    }
                    else {
                        resolve(parseStudyJson(data, this._cache.log));
                    }
                }
            });
        });
    }
    /**
     * Remove the entry from the cache. All this does is delete the underlying file. The returned Promise is more for
     * error handling than anything else.
     */
    remove() {
        return new Promise((resolve, reject) => {
            fs.unlink(this.filename, (err) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve();
                }
            });
        });
    }
}
exports.CacheEntry = CacheEntry;
/**
 * Create a directory if it does not already exist. If it does exist (or some other file exists in its place), this
 * still resolves rather than raising an error. This used to check if the paths were really directories, but since the
 * final step is now reading the file entries from the directory anyway, that check ensures everything is a directory
 * anyway.
 *
 * @param path the path to create (will be created recursively)
 * @return a Promise that resolves as true if the directory was newly created or false if it existed
 */
function mkdir(fs, path) {
    return new Promise((resolve, reject) => {
        fs.mkdir(path, (err) => {
            if (err) {
                if (err.code === 'EEXIST') {
                    // This is fine - resolve false. We'll only get this if the final part of the path exists (although it will be
                    // EEXIST regardless of if it's a directory or a regular file)
                    resolve(false);
                }
                else {
                    reject(err);
                }
            }
            else {
                resolve(true);
            }
        });
    });
}
exports.mkdir = mkdir;
/**
 * System to fill in data on research studies based on the trial data from
 * https://clinicaltrials.gov/
 *
 * The current system works as follows:
 *
 * Call updateResearchStudies() with the ResearchStudy objects to update. It will locate all ResearchStudy objects that
 * contain an NCT ID. It will then:
 *
 * 1. Call downloadTrials() to download the ClinicalStudy data from clinicaltrials.gov. It starts an HTTPS request to
 *    download the trial results and, if successful, passes the response stream off to extractResults().
 * 2. extractResults() saves the stream its given to a temporary ZIP file and extracts that file to a temporary
 *    directory. Once the ZIP is extracted, it deletes the temporary ZIP. It then returns the directory where the files
 *    were extracted.
 * 3. With the path to the extracted ZIP, updateResearchStudies() can now use the downloaded trial data to update the
 *    trials.
 * 4. Before resolving, the temporary directory is deleted.
 *
 * A future version will likely change this to make it so that rather than immediately deleting the temporary trial
 * data, it's kept around in a least-recently-used cache and only deleted as the cache exceeds a given size. This will
 * add a few new public APIs to enable the retrieval of clinical study data by NCT ID. Right now, though, there is no
 * way to extract the clinical study data directly.
 */
class ClinicalTrialsGovService {
    /**
     * The maximum number of trials to attempt to download in a single request. Each NCT ID increases the URI size by
     * effectively 12 bytes: essentially each NCT ID becomes ',NCT12345678'. The request path "counts" as a header and
     * has to be smaller than maxHeaderSize given to the client. By default, this uses the default client, so the client
     * should be the default HTTPS client. In theory this would support around 500 per request, instead this defaults to
     * 128 which is a nice round number. (In binary, at least.)
     */
    get maxTrialsPerRequest() {
        return this._maxTrialsPerRequest;
    }
    set maxTrialsPerRequest(newValue) {
        // Basically, just make sure the new value is positive and an integer. Positive infinity is fine and remains
        // positive infinity: it means never split.
        if (newValue > 1) {
            this._maxTrialsPerRequest = Math.floor(newValue);
        }
    }
    /**
     * The expiration timeout, the time in milliseconds after a cache entry should be considered expired. The default
     * is 1 hour.
     */
    get expirationTimeout() {
        return this._expirationTimeout;
    }
    set expirationTimeout(value) {
        // Ensure it's at least 1000ms
        this._expirationTimeout = Math.max(value, 1000);
    }
    /**
     * The interval between cleanup sweeps. Set to 0 (or any negative value) or Infinity to disable periodic cleanup. Note
     * that the value is limited to the range 60000 (a minute)-2147483647 (the maximum value of a signed 32-bit integer)
     * as that maximum value is the maximum allowed timeout for a timer within Node.js. If the value is set to a negative
     * number or Infinity, the retrieved value will be 0, to indicate that the interval will not be set.
     */
    get cleanupInterval() {
        return this._cleanupIntervalMillis;
    }
    set cleanupInterval(value) {
        // Clamp it to 0 (special value for "never") or at least 60000 - also clamp the maximum to a 32-bit signed integer
        // which is the maximum allowed timeout within Node.js. (Exceeding that value changes the timeout to 1ms. Surprise!
        // I mean, sure, that's documented, but...)
        this._cleanupIntervalMillis =
            value <= 0 ? 0 : value === Infinity ? 0 : Math.min(Math.max(Math.floor(value), 60000), 0x7fffffff);
    }
    /**
     * Creates a new instance. This will not initialize the cache or load anything, this merely creates the object. Use
     * the #init() method to initialize the service and load existing data. (Or use createClinicalTrialsGovService() do
     * construct and initialize at the same time.)
     *
     * @param dataDir the directory to use for cache data
     * @param log an optional function to receive debug log messages. By default this uses util.debuglog('ctgovservice')
     *     meaning that the log can be activated by setting NODE_DEBUG to "ctgovservice"
     */
    constructor(dataDir, options) {
        var _a, _b, _c;
        this.dataDir = dataDir;
        /**
         * Internal value to track temporary file names.
         */
        this.tempId = 0;
        this._maxTrialsPerRequest = 128;
        this._expirationTimeout = 60 * 60 * 1000;
        this._cleanupIntervalMillis = 60 * 60 * 1000;
        /**
         * The maximum allowed entry size before it is rejected. Default is 128MB, which is hopefully well more than necessary
         * for any reasonable clinical trial JSON.
         */
        this.maxAllowedEntrySize = 128 * 1024 * 1024;
        /**
         * Actual cache of NCT IDs to their cached values. This is the "real cache" versus whatever's on the filesystem.
         * (The NCT ID is in the form that returns true from isValidNCTNumber, so NCTnnnnnnnn where n are digits.)
         */
        this.cache = new Map();
        this.cleanupTimeout = null;
        this.cacheDataDir = path.join(dataDir, 'data');
        const log = options ? options.log : undefined;
        // If no log was given, create it
        this.log = log !== null && log !== void 0 ? log : (0, util_1.debuglog)('ctgovservice');
        // Default expiration timeout to an hour
        this.expirationTimeout = (_a = options === null || options === void 0 ? void 0 : options.expireAfter) !== null && _a !== void 0 ? _a : 60 * 60 * 1000;
        // Default cleanup interval to an hour
        this.cleanupInterval = (_b = options === null || options === void 0 ? void 0 : options.cleanInterval) !== null && _b !== void 0 ? _b : 60 * 60 * 1000;
        this.fs = (_c = options === null || options === void 0 ? void 0 : options.fs) !== null && _c !== void 0 ? _c : fs;
        this.service = new clinicaltrialsgov_api_1.ClinicalTrialsGovAPI({ logger: this.log });
    }
    /**
     * Creates the necessary directories if they do not exist, and loads any existing data into the cache. Parent
     * directories of the cache will not be created automatically, they must already exist.
     */
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            this.log('Using %s as cache dir for clinicaltrials.gov data', this.dataDir);
            const baseDirExisted = !(yield mkdir(this.fs, this.dataDir));
            // The directory containing cached studies
            const dataDirExisted = !(yield mkdir(this.fs, this.cacheDataDir));
            if (baseDirExisted && dataDirExisted) {
                // If both directories existed, it's necessary to restore the cache directory
                yield this.restoreCacheFromFS();
                this.log('Restored existing cache data.');
            }
            else {
                this.log(baseDirExisted ? 'Created data directory for storing result' : 'Created new cache directory');
            }
            // Once started, run the cache cleanup every _cleanupIntervalMillis
            this.setCleanupTimeout();
        });
    }
    /**
     * Sets the timeout interval assuming it was set.
     */
    setCleanupTimeout() {
        // If set to infinity we internally set it to 0
        if (this._cleanupIntervalMillis > 0) {
            this.cleanupTimeout = setTimeout(() => {
                this.removeExpiredCacheEntries().then(() => {
                    // Set up to do this again when that's done
                    this.setCleanupTimeout();
                }, (error) => {
                    this.log('Error cleaning expired cache entries: %o', error);
                    // Even though this attempt failed, try again later
                    this.setCleanupTimeout();
                });
            }, this._cleanupIntervalMillis);
        }
    }
    /**
     * Shuts down the service, doing any final necessary cleanup. At present this stops any running timers. The Promise
     * returned currently resolves immediately, but in the future, it may resolve asynchronously when the service has been
     * cleanly shut down.
     */
    destroy() {
        if (this.cleanupTimeout !== null) {
            clearTimeout(this.cleanupTimeout);
            // And blank it
            this.cleanupTimeout = null;
        }
        return Promise.resolve();
    }
    /**
     * This attempts to load the cache from the filesystem.
     */
    restoreCacheFromFS() {
        return new Promise((resolve, reject) => {
            this.log('Scanning %s for existing cache entries...', this.cacheDataDir);
            this.fs.readdir(this.cacheDataDir, (err, files) => {
                if (err) {
                    reject(err);
                }
                else {
                    // Go through the files and create entries for them.
                    const promises = [];
                    for (const file of files) {
                        this.log('Checking %s', file);
                        // Split this file name into two parts: the extension and the base name
                        const dotIdx = file.lastIndexOf('.');
                        if (dotIdx < 1) {
                            // Skip "bad" files
                            continue;
                        }
                        // FIXME: This is probably a bad idea. Right now the file name serves as the "master" name for files. It's
                        // probably possible to load the file and pull the NCT ID out that way, as well as clear out files that
                        // can't be parsed.
                        const baseName = file.substring(0, dotIdx);
                        const extension = file.substring(dotIdx + 1);
                        if (isValidNCTNumber(baseName) && extension === 'json') {
                            promises.push(this.createCacheEntry(baseName, path.join(this.cacheDataDir, file)));
                        }
                    }
                    Promise.all(promises).then(() => {
                        resolve();
                    }, reject);
                }
            });
        });
    }
    createCacheEntry(id, filename) {
        // Restoring this involves getting the time the file was created so we know when the entry expires
        return new Promise((resolve, reject) => {
            this.fs.stat(filename, (err, stats) => {
                if (err) {
                    // TODO (maybe): Instead of rejecting, just log - rejecting will caused Promise.all to immediately reject and
                    // ignore the rest of the Promises. However, stat failing is probably a "real" error.
                    reject(err);
                }
                else {
                    this.cache.set(id, new CacheEntry(this, filename, { stats: stats }));
                    this.log('Restored cache entry for %s', filename);
                    resolve();
                }
            });
        });
    }
    /**
     * Attempts to update the given research studies with data from within this store. This returns a Promise that
     * resolves when the research studies have been updated with any available additional data. The Promise may be
     * rejected if the process fails entirely (that is, clinicaltrials.gov could not be contacted at all, or an I/O error
     * occurs).
     *
     * @param studies the studies to attempt to update
     * @returns a Promise that resolves when the studies are updated. It will resolve with the same array that was passed
     *     in - this updates the given objects, it does not clone them and create new ones.
     */
    updateResearchStudies(studies) {
        const nctIdMap = findNCTNumbers(studies);
        if (nctIdMap.size === 0) {
            // Nothing to do
            return Promise.resolve(studies);
        }
        else {
            const nctIds = Array.from(nctIdMap.keys());
            // Make sure the NCT numbers are in the cache
            return this.ensureTrialsAvailable(nctIds).then(() => {
                const promises = [];
                // Go through the NCT numbers we found and updated all matching trials
                for (const entry of nctIdMap.entries()) {
                    const [nctId, study] = entry;
                    const promise = this.getCachedClinicalStudy(nctId);
                    promises.push(promise);
                    promise.then((clinicalStudy) => {
                        if (clinicalStudy !== null) {
                            // Make sure we have data to use - we may still get null if there was no data for a given NCT number
                            // Update whatever trials we have
                            if (Array.isArray(study)) {
                                for (const s of study) {
                                    this.updateResearchStudy(s, clinicalStudy);
                                }
                            }
                            else {
                                this.updateResearchStudy(study, clinicalStudy);
                            }
                        }
                    });
                }
                // Finally resolve to the promises we were given
                return Promise.all(promises).then(() => studies);
            });
        }
    }
    /**
     * Tells the cache to delete all expired cached files. The entries are removed from access immediately, but a Promise
     * is returned that resolves when all the underlying data is cleaned up. This Promise is more for error handling than
     * anything else.
     */
    removeExpiredCacheEntries() {
        const expiredEntries = [];
        const expiredIds = [];
        // Go through the cache and find all expired entries
        const expired = new Date();
        expired.setTime(expired.getTime() - this._expirationTimeout);
        for (const [key, entry] of this.cache.entries()) {
            if (entry.lastAccessedBefore(expired)) {
                expiredIds.push(key);
                expiredEntries.push(entry);
            }
        }
        const expiredPromises = [];
        // Now that the expired entries have been found, remove them
        for (const key of expiredIds) {
            this.cache.delete(key);
        }
        // FIXME: It's unclear whether or not this creates a race condition where it may be possible for another request
        // for an entry we just expired to be deleted before the new entry is ready. This is somewhat unlikely, but it's
        // something that could probably be fixed by ensuring that each time we create a cache entry we use a unique
        // filename.
        for (const entry of expiredEntries) {
            expiredPromises.push(entry.remove());
        }
        // The "then" essentially removes the array of undefined that will result
        // into a single undefined.
        return Promise.all(expiredPromises).then(() => {
            /* do nothing */
        });
    }
    ensureTrialsAvailable(idsOrStudies) {
        // We only want string IDs and we may end up filtering some of them out
        const ids = [];
        for (const o of idsOrStudies) {
            if (typeof o === 'string') {
                if (isValidNCTNumber(o))
                    ids.push(o);
            }
            else {
                const id = findNCTNumber(o);
                if (id)
                    ids.push(id);
            }
        }
        // Now that we have the IDs, we can split them into download requests
        const promises = [];
        for (let start = 0; start < ids.length; start += this.maxTrialsPerRequest) {
            promises.push(this.downloadTrials(ids.slice(start, Math.min(start + this.maxTrialsPerRequest, ids.length))));
        }
        return Promise.all(promises).then(() => {
            // This exists solely to turn the result from an array of success/fails into a single nothing
        });
    }
    /**
     * Downloads the given trials from ClinicalTrials.gov and stores them in the data directory. Note that this will
     * always replace trials if they exist.
     *
     * @param ids the IDs of the trials to download
     * @returns true if the request succeeded, false if it failed for some reason - the exact reason can be logged but
     * is otherwise silently eaten
     */
    downloadTrials(ids) {
        return __awaiter(this, void 0, void 0, function* () {
            let success = false;
            // Now that we're starting to download clinical trials, immediately create pending entries for them.
            for (const id of ids) {
                if (!this.cache.has(id)) {
                    this.cache.set(id, new CacheEntry(this, this.pathForNctNumber(id), { pending: true }));
                }
            }
            try {
                const studies = yield this.service.fetchStudies(ids, this._maxTrialsPerRequest);
                for (const study of studies) {
                    yield this.addCacheEntry(study);
                }
                success = true;
            }
            catch (ex) {
                // If an error occurred while fetching studies, every cache entry we just loaded may be invalid.
                this.log('Error while fetching trials: %o', ex);
                this.log('Invalidating cache entry IDs for: %s', ids);
                for (const id of ids) {
                    const entry = this.cache.get(id);
                    if (entry && entry.pending) {
                        this.cache.delete(id);
                    }
                }
            }
            // Invalidate any IDs that are still pending, they weren't in the results
            for (const id of ids) {
                const entry = this.cache.get(id);
                if (entry && entry.createdAt === null) {
                    this.log('Removing cache entry for %s: it was not in the downloaded bundle!', id);
                    entry.fail('Not found in bundle');
                    this.cache.delete(id);
                }
            }
            return success;
        });
    }
    /**
     * Create a path to the data file that stores data about a cache entry.
     * @param nctNumber the NCT number
     */
    pathForNctNumber(nctNumber) {
        // FIXME: It's probably best not to use the NCT number as the sole part of the filename. See
        // removeExpiredCacheEntries for details.
        return path.join(this.cacheDataDir, nctNumber + '.json');
    }
    addCacheEntry(study) {
        var _a, _b;
        // See if we can locate an NCT number for this study.
        const nctNumber = (_b = (_a = study.protocolSection) === null || _a === void 0 ? void 0 : _a.identificationModule) === null || _b === void 0 ? void 0 : _b.nctId;
        if (typeof nctNumber !== 'string') {
            this.log('Ignoring study object from server: unable to locate an NCT ID for it! (protocolSection.identificationModule.nctId missing or not a string)');
            return Promise.resolve();
        }
        const filename = this.pathForNctNumber(nctNumber);
        // The cache entry should already exist
        const entry = this.cache.get(nctNumber);
        // Tell the entry that we are writing data
        if (entry) {
            entry.found();
        }
        else {
            this.log('No cache entry for %s! NOT writing it to cache! (Got back a different NCT number?)', nctNumber);
            return Promise.resolve();
        }
        return new Promise((resolve, reject) => {
            this.fs.writeFile(filename, JSON.stringify(study), 'utf8', (err) => {
                if (err) {
                    this.log('Unable to create file [%s]: %o', filename, err);
                    // If the cache entry exists in pending mode, delete it - we failed to create this entry
                    if (entry && entry.pending) {
                        entry.fail('Unable to create file');
                        this.cache.delete(nctNumber);
                    }
                    // TODO: Do we also need to delete the file? Or will the error prevent the file from existing?
                    reject(err);
                }
                else {
                    if (entry && entry.pending) {
                        entry.ready();
                    }
                    resolve();
                }
            });
        });
    }
    /**
     * Loads a ClinicalStudy from an extracted dataset. This will never download a copy, this will only ever return from
     * within the cache.
     * @param nctNumber the NCT number
     * @returns a Promise that resolves to either the parsed ClinicalStudy or to null if the ClinicalStudy does not exist
     */
    getCachedClinicalStudy(nctNumber) {
        const entry = this.cache.get(nctNumber);
        if (entry) {
            return entry.load();
        }
        else {
            return Promise.resolve(null);
        }
    }
    /**
     * The provides a stub that handles updating the research study with data from a clinical study downloaded from the
     * ClinicalTrials.gov website. This primarily exists as a stub to allow the exact process which updates a research
     * study to be overridden if necessary.
     *
     * @param researchStudy the research study to update
     * @param clinicalStudy the clinical study to update it with
     */
    updateResearchStudy(researchStudy, clinicalStudy) {
        (0, study_fhir_converter_1.updateResearchStudyWithClinicalStudy)(researchStudy, clinicalStudy);
    }
    /**
     * Creates and initializes a new service for retrieving data from http://clinicaltrials.gov/. This will automatically
     * invoke the init method to create the directory if it doesn't exist and load any existing data if it does. Note that
     * the init method will not create missing parent directories - the path to the cache directory must already exist
     * minus the cache directory itself. If the cache directory cannot be created the Promise will be rejected with the
     * error preventing it from being created.
     *
     * @param dataDir the data directory
     * @param options additional options that can be set to further configure the trial service
     * @returns a Promise that resolves when the service is ready
     */
    static create(dataDir, options) {
        const result = new ClinicalTrialsGovService(dataDir, options);
        return result.init().then(() => result);
    }
}
exports.ClinicalTrialsGovService = ClinicalTrialsGovService;
/**
 * Creates and initializes a new service for retrieving data from http://clinicaltrials.gov/. This is the same as
 * ClinicalTrialsGovService.create, see that method for details.
 *
 * @param dataDir the data directory
 * @param options additional options that can be set to further configure the trial service
 * @returns a Promise that resolves when the service is ready
 */
function createClinicalTrialsGovService(dataDir, options) {
    return ClinicalTrialsGovService.create(dataDir, options);
}
exports.createClinicalTrialsGovService = createClinicalTrialsGovService;
