"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.mCODEextractor = exports.extractCoding = void 0;
const mcode_1 = require("./mcode");
const fhir_util_1 = require("./fhir-util");
/**
 * Extract coding from a codeable concept. Exported to allow testing.
 * @param codeableConcept the codeable concept to check
 * @return either the coding values or an empty array if not found
 */
function extractCoding(codeableConcept) {
    if (codeableConcept) {
        if (Array.isArray(codeableConcept)) {
            // This is basically an overload, but in this case, we want to concatenate
            // all the codings together into one array
            const coding = [];
            for (const concept of codeableConcept) {
                if (Array.isArray(concept.coding)) {
                    coding.push(...concept.coding);
                }
            }
            return coding;
        }
        else {
            const coding = codeableConcept.coding;
            if (Array.isArray(coding)) {
                return coding;
            }
        }
    }
    return [];
}
exports.extractCoding = extractCoding;
/**
 * Class that extracts mCODE data from a patient record.
 */
class mCODEextractor {
    /**
     * Constructor.
     * @param patientBundle The patient bundle to build the mCODE mapping from.
     */
    constructor(patientBundle) {
        var _a, _b, _c, _d, _e, _f;
        /**
         * Extracted ECOG performance status, or -1 if not found.
         */
        this.ecogPerformanceStatus = -1;
        /**
         * Extracted Karnofsky performance status, or -1 if not found.
         */
        this.karnofskyPerformanceStatus = -1;
        this.primaryCancerConditions = [];
        this.TNMClinicalStageGroups = [];
        this.TNMPathologicalStageGroups = [];
        this.secondaryCancerConditions = [];
        this.tumorMarkers = [];
        this.cancerRelatedRadiationProcedures = [];
        this.cancerRelatedSurgicalProcedures = [];
        // Temporary store of related reasons
        const cancerRelatedSurgialProceduresReferences = [];
        this.cancerRelatedMedicationStatements = [];
        this.cancerGeneticVariants = [];
        this.birthDate = null;
        if (!patientBundle || !Array.isArray(patientBundle.entry)) {
            throw Error('Input Patient Bundle is invalid/has no entries');
        }
        else {
            for (const entry of patientBundle.entry) {
                const resource = entry.resource;
                if (typeof resource !== 'object' || resource === null) {
                    // Skip invalid resources
                    continue;
                }
                if (resource.resourceType === 'Condition') {
                    if ((0, fhir_util_1.resourceContainsProfile)(resource, mcode_1.MCODE_PRIMARY_CANCER_CONDITION)) {
                        const tempPrimaryCancerCondition = {
                            clinicalStatus: extractCoding(resource.clinicalStatus),
                            meta_profile: 'mcode-primary-cancer-condition',
                            histologyMorphologyBehavior: [],
                            coding: extractCoding(resource.code),
                            id: resource.id
                        };
                        const extensions = resource.extension;
                        if (Array.isArray(extensions)) {
                            for (const extension of extensions) {
                                if (extension.url === mcode_1.MCODE_HISTOLOGY_MORPHOLOGY_BEHAVIOR) {
                                    const coding = (_a = extension.valueCodeableConcept) === null || _a === void 0 ? void 0 : _a.coding;
                                    if (Array.isArray(coding)) {
                                        tempPrimaryCancerCondition.histologyMorphologyBehavior = coding;
                                    }
                                }
                            }
                        }
                        this.primaryCancerConditions.push(tempPrimaryCancerCondition);
                    }
                    if ((0, fhir_util_1.resourceContainsProfile)(resource, mcode_1.MCODE_SECONDARY_CANCER_CONDITION)) {
                        const tempSecondaryCancerCondition = {
                            clinicalStatus: extractCoding(resource.clinicalStatus),
                            meta_profile: 'mcode-secondary-cancer-condition',
                            id: resource.id,
                            bodySite: extractCoding(resource.bodySite),
                            coding: extractCoding(resource.code)
                        };
                        this.secondaryCancerConditions.push(tempSecondaryCancerCondition); // needs specific de-dup helper function
                    }
                }
                if (resource.resourceType === 'Observation') {
                    if ((0, fhir_util_1.resourceContainsProfile)(resource, mcode_1.MCODE_TNM_CLINICAL_STAGE_GROUP)) {
                        this.TNMClinicalStageGroups = this.addCoding(this.TNMClinicalStageGroups, (_b = resource.valueCodeableConcept) === null || _b === void 0 ? void 0 : _b.coding);
                    }
                    if ((0, fhir_util_1.resourceContainsProfile)(resource, mcode_1.MCODE_TNM_PATHOLOGICAL_STAGE_GROUP)) {
                        this.TNMPathologicalStageGroups = this.addCoding(this.TNMPathologicalStageGroups, (_c = resource.valueCodeableConcept) === null || _c === void 0 ? void 0 : _c.coding);
                    }
                    if ((0, fhir_util_1.resourceContainsProfile)(resource, mcode_1.MCODE_TUMOR_MARKER)) {
                        const valueQuantity = resource.valueQuantity;
                        const valueRatio = resource.valueRatio;
                        // For now, ignore values that are missing both a quantity and ratio
                        if (valueQuantity !== undefined || valueRatio !== undefined) {
                            this.tumorMarkers.push({
                                valueQuantity: valueQuantity,
                                valueRatio: valueRatio,
                                valueCodeableConcept: extractCoding(resource.valueCodeableConcept),
                                interpretation: extractCoding(resource.interpretation),
                                coding: extractCoding(resource.code)
                            });
                        }
                    }
                    // Parse and Extract mCODE Cancer Genetic Variant
                    if ((0, fhir_util_1.resourceContainsProfile)(resource, mcode_1.MCODE_CANCER_GENETIC_VARIANT)) {
                        const tempCGV = {
                            coding: extractCoding(resource.code),
                            component: {
                                geneStudied: [],
                                genomicsSourceClass: []
                            },
                            valueCodeableConcept: [],
                            interpretation: []
                        };
                        const components = resource.component;
                        if (Array.isArray(components)) {
                            for (const currentComponent of components) {
                                const coding = (_d = currentComponent.code) === null || _d === void 0 ? void 0 : _d.coding;
                                if (Array.isArray(coding)) {
                                    for (const currentComponentCode of coding) {
                                        if (currentComponentCode.code == '48018-6') {
                                            // With this code, we've reached a GeneStudied. Populate the GeneStudied attribute.
                                            tempCGV.component.geneStudied.push(currentComponent);
                                        }
                                        if (currentComponentCode.code == '48002-0') {
                                            // With this code, we've reached a GenomicSourceClass. Populate the GenomicSourceClass attribute.
                                            tempCGV.component.genomicsSourceClass.push(currentComponent);
                                        }
                                    }
                                }
                            }
                        }
                        tempCGV.valueCodeableConcept = extractCoding(resource.valueCodeableConcept);
                        tempCGV.interpretation = extractCoding(resource.interpretation);
                        this.cancerGeneticVariants.push(tempCGV);
                    }
                    if ((0, fhir_util_1.resourceContainsProfile)(resource, mcode_1.MCODE_ECOG_PERFORMANCE_STATUS)) {
                        if (typeof resource.valueInteger === 'number')
                            this.ecogPerformanceStatus = resource.valueInteger;
                    }
                    if ((0, fhir_util_1.resourceContainsProfile)(resource, mcode_1.MCODE_KARNOFSKY_PERFORMANCE_STATUS)) {
                        if (typeof resource.valueInteger === 'number')
                            this.karnofskyPerformanceStatus = resource.valueInteger;
                    }
                }
                if (resource.resourceType === 'Patient' &&
                    // FIXME: Really constrain to resources with this profile? This does not appear to be set in the app!
                    (0, fhir_util_1.resourceContainsProfile)(resource, mcode_1.MCODE_CANCER_PATIENT)) {
                    if (resource.birthDate) {
                        this.birthDate = resource.birthDate;
                    }
                }
                if (resource.resourceType === 'Procedure') {
                    if ((0, fhir_util_1.resourceContainsProfile)(resource, mcode_1.MCODE_CANCER_RELATED_RADIATION_PROCEDURE)) {
                        const tempCancerRelatedRadiationProcedure = {
                            bodySite: extractCoding(resource.bodySite),
                            mcodeTreatmentIntent: [],
                            coding: extractCoding(resource.code)
                        };
                        if (!this.listContainsProcedure(this.cancerRelatedRadiationProcedures, tempCancerRelatedRadiationProcedure)) {
                            this.cancerRelatedRadiationProcedures.push(tempCancerRelatedRadiationProcedure);
                        }
                    }
                    if ((0, fhir_util_1.resourceContainsProfile)(resource, mcode_1.MCODE_CANCER_RELATED_SURGICAL_PROCEDURE)) {
                        const tempCancerRelatedSurgicalProcedure = {
                            bodySite: extractCoding(resource.bodySite),
                            reasonReference: [],
                            coding: extractCoding(resource.code)
                        };
                        if (!this.listContainsProcedure(this.cancerRelatedSurgicalProcedures, tempCancerRelatedSurgicalProcedure)) {
                            this.cancerRelatedSurgicalProcedures.push(tempCancerRelatedSurgicalProcedure);
                            cancerRelatedSurgialProceduresReferences.push((_e = resource.reasonReference) !== null && _e !== void 0 ? _e : []);
                        }
                    }
                }
                if (resource.resourceType === 'MedicationStatement' &&
                    (0, fhir_util_1.resourceContainsProfile)(resource, mcode_1.MCODE_CANCER_RELATED_MEDICATION_STATEMENT)) {
                    this.cancerRelatedMedicationStatements = this.addCoding(this.cancerRelatedMedicationStatements, (_f = resource.medicationCodeableConcept) === null || _f === void 0 ? void 0 : _f.coding);
                }
            }
        }
        // Checking if the performanceStatus exists and also making sure it's not 0, as 0 is a valid score
        if (!this.ecogPerformanceStatus && this.ecogPerformanceStatus != 0) {
            this.ecogPerformanceStatus = -1;
        }
        if (!this.karnofskyPerformanceStatus && this.karnofskyPerformanceStatus != 0) {
            this.karnofskyPerformanceStatus = -1;
        }
        // Once all resources are loaded, check to add the meta.profile for cancer related surgical procedure reason references.
        for (let idx = 0; idx < this.cancerRelatedSurgicalProcedures.length; idx++) {
            const procedure = this.cancerRelatedSurgicalProcedures[idx];
            const conditions = this.primaryCancerConditions.concat(this.secondaryCancerConditions);
            // Populate the reasonReference field
            for (const reasonReference of cancerRelatedSurgialProceduresReferences[idx]) {
                for (const condition of conditions) {
                    if (condition.id === reasonReference.reference) {
                        procedure.reasonReference.push({
                            reference: reasonReference.reference,
                            display: reasonReference.display,
                            meta_profile: condition.meta_profile
                        });
                        // Don't continue looking for matching resources
                        break;
                    }
                }
            }
        }
    }
    /**
     * Checks whether the given list contains the given coding.
     * @param codingList
     * @param coding
     * @returns
     */
    contains(codingList, coding) {
        return codingList.some((list_coding) => list_coding.system === coding.system && list_coding.code === coding.code);
    }
    /**
     * Adds the given coding to the given coding list.
     * @param codingList
     * @param codes the codes to add - if undefined, nothing is added (undefined
     *    is allowed to make checking easier)
     * @returns
     */
    addCoding(codingList, codes) {
        if (Array.isArray(codes)) {
            for (const code of codes) {
                if (!this.contains(codingList, code)) {
                    codingList.push(code);
                }
            }
        }
        return codingList;
    }
    /**
     * Returns whether the given list contains the given procedure.
     * @param procedureList
     * @param procedure
     * @returns
     */
    listContainsProcedure(procedureList, procedure) {
        for (const storedProcedure of procedureList) {
            if (procedure.coding.every((coding1) => storedProcedure.coding.some((coding2) => coding1.system == coding2.system && coding1.code == coding2.code)) &&
                (!procedure.bodySite ||
                    !storedProcedure.bodySite ||
                    procedure.bodySite.every((coding1) => storedProcedure.coding.some((coding2) => coding1.system == coding2.system && coding1.code == coding2.code)))) {
                return true;
            }
        }
        return false;
    }
}
exports.mCODEextractor = mCODEextractor;
