"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.updateResearchStudyWithClinicalStudy = exports.convertRecruitmentStatusToLocationStatus = exports.convertClincalStudyStatusToFHIRStatus = exports.convertToResearchStudyPhase = void 0;
const ctg_api_1 = require("./ctg-api");
const fhir_type_guards_1 = require("./fhir-type-guards");
const research_study_1 = require("./research-study");
const PHASE_MAP = new Map([
    [ctg_api_1.Phase.NA, 'n-a'],
    [ctg_api_1.Phase.EARLY_PHASE1, 'early-phase-1'],
    [ctg_api_1.Phase.PHASE1, 'phase-1'],
    [ctg_api_1.Phase.PHASE2, 'phase-2'],
    [ctg_api_1.Phase.PHASE3, 'phase-3'],
    [ctg_api_1.Phase.PHASE4, 'phase-4']
]);
function convertToResearchStudyPhase(phase) {
    return PHASE_MAP.get(phase);
}
exports.convertToResearchStudyPhase = convertToResearchStudyPhase;
const CLINICAL_STATUS_MAP = new Map([
    [ctg_api_1.Status.ACTIVE_NOT_RECRUITING, 'closed-to-accrual'],
    [ctg_api_1.Status.COMPLETED, 'completed'],
    // FIXME: This does not appear to have a proper mapping
    [ctg_api_1.Status.ENROLLING_BY_INVITATION, 'active'],
    [ctg_api_1.Status.NOT_YET_RECRUITING, 'approved'],
    [ctg_api_1.Status.RECRUITING, 'active'],
    [ctg_api_1.Status.SUSPENDED, 'temporarily-closed-to-accrual'],
    [ctg_api_1.Status.TERMINATED, 'administratively-completed'],
    [ctg_api_1.Status.WITHDRAWN, 'withdrawn'],
    [ctg_api_1.Status.AVAILABLE, 'completed'],
    [ctg_api_1.Status.NO_LONGER_AVAILABLE, 'closed-to-accrual'],
    [ctg_api_1.Status.TEMPORARILY_NOT_AVAILABLE, 'temporarily-closed-to-accrual'],
    [ctg_api_1.Status.APPROVED_FOR_MARKETING, 'completed'],
    // FIXME: This does not appear to have a proper mapping
    [ctg_api_1.Status.WITHHELD, 'in-review'],
    // FIXME: This does not appear to have a proper mapping
    [ctg_api_1.Status.UNKNOWN, 'in-review']
]);
function convertClincalStudyStatusToFHIRStatus(status) {
    return CLINICAL_STATUS_MAP.get(status);
}
exports.convertClincalStudyStatusToFHIRStatus = convertClincalStudyStatusToFHIRStatus;
const LOCATION_STATUS_MAP = new Map([
    [ctg_api_1.RecruitmentStatus.ACTIVE_NOT_RECRUITING, 'active'],
    [ctg_api_1.RecruitmentStatus.COMPLETED, 'inactive'],
    [ctg_api_1.RecruitmentStatus.ENROLLING_BY_INVITATION, 'active'],
    [ctg_api_1.RecruitmentStatus.NOT_YET_RECRUITING, 'inactive'],
    [ctg_api_1.RecruitmentStatus.RECRUITING, 'active'],
    [ctg_api_1.RecruitmentStatus.SUSPENDED, 'suspended'],
    [ctg_api_1.RecruitmentStatus.TERMINATED, 'inactive'],
    [ctg_api_1.RecruitmentStatus.WITHDRAWN, 'inactive'],
    [ctg_api_1.RecruitmentStatus.AVAILABLE, 'active']
]);
function convertRecruitmentStatusToLocationStatus(status) {
    return LOCATION_STATUS_MAP.get(status);
}
exports.convertRecruitmentStatusToLocationStatus = convertRecruitmentStatusToLocationStatus;
function convertToTitleCase(str) {
    return str
        .replace(/([A-Z]+)/g, (s) => (s.length > 1 ? s.substring(0, 1) + s.substring(1).toLowerCase() : s))
        .replace(/_/g, ' ');
}
function convertArrayToCodeableConcept(trialConditions) {
    const fhirConditions = [];
    for (const condition of trialConditions) {
        fhirConditions.push({ text: condition });
    }
    return fhirConditions;
}
/**
 * Updates a research study with data from a clinical study off the ClinicalTrials.gov website. This will only update
 * fields that do not have data, it will not overwrite any existing data.
 *
 * Mapping as defined by https://www.hl7.org/fhir/researchstudy-mappings.html#clinicaltrials-gov
 *
 * @param result the research study to update
 * @param study the clinical study to use to update
 */
function updateResearchStudyWithClinicalStudy(result, study) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r;
    const protocolSection = study.protocolSection;
    // If there is no protocol section, we can't do anything.
    if (!protocolSection) {
        return result;
    }
    if (!result.enrollment) {
        const eligibility = protocolSection.eligibilityModule;
        if (eligibility) {
            const criteria = eligibility.eligibilityCriteria;
            if (criteria) {
                const group = { resourceType: 'Group', id: 'group' + result.id, type: 'person', actual: false };
                const reference = (0, research_study_1.addContainedResource)(result, group);
                reference.display = criteria;
                result.enrollment = [reference];
            }
        }
    }
    if (!result.description) {
        const briefSummary = (_a = protocolSection.descriptionModule) === null || _a === void 0 ? void 0 : _a.briefSummary;
        if (briefSummary) {
            result.description = briefSummary;
        }
    }
    if (!result.phase) {
        const phase = (_b = protocolSection.designModule) === null || _b === void 0 ? void 0 : _b.phases;
        if (phase && phase.length > 0) {
            // For now, just grab whatever the first phase is
            // TODO: handle the somewhat weirder "phase-1-phase-2" items
            const code = convertToResearchStudyPhase(phase[0]);
            if (code) {
                const display = code.replace(/-/g, ' ');
                result.phase = {
                    coding: [
                        {
                            system: 'http://terminology.hl7.org/CodeSystem/research-study-phase',
                            code: code,
                            display: display
                        }
                    ],
                    text: display
                };
            }
        }
    }
    // ------- Category
    // Since we may not have all of the Study design in the result, we need to do a merge of data
    const categories = result.category ? result.category : [];
    // We need to determine what categories have already been declared.
    const types = categories.map((item) => {
        var _a;
        const sep = (_a = item.text) === null || _a === void 0 ? void 0 : _a.split(':');
        return sep ? sep[0] : '';
    });
    const studyType = (_d = (_c = study.protocolSection) === null || _c === void 0 ? void 0 : _c.designModule) === null || _d === void 0 ? void 0 : _d.studyType;
    if (studyType && !types.includes('Study Type')) {
        categories.push({ text: 'Study Type: ' + convertToTitleCase(studyType) });
    }
    const designInfo = (_e = protocolSection.designModule) === null || _e === void 0 ? void 0 : _e.designInfo;
    if (designInfo) {
        if (!types.includes('Intervention Model')) {
            if (designInfo.interventionModel) {
                categories.push({ text: 'Intervention Model: ' + convertToTitleCase(designInfo.interventionModel) });
            }
            else if (designInfo.interventionModelDescription) {
                categories.push({ text: 'Intervention Model: ' + designInfo.interventionModelDescription });
            }
        }
        if (designInfo.primaryPurpose && !types.includes('Primary Purpose')) {
            categories.push({ text: 'Primary Purpose: ' + convertToTitleCase(designInfo.primaryPurpose) });
        }
        const maskingInfo = designInfo.maskingInfo;
        if (maskingInfo && !types.includes('Masking')) {
            // It's unclear exactly how to convert this
            const masking = maskingInfo.masking ? convertToTitleCase(maskingInfo.masking) : maskingInfo.maskingDescription;
            if (masking) {
                categories.push({ text: 'Masking: ' + masking });
            }
        }
        if (designInfo.allocation && !types.includes('Allocation')) {
            categories.push({ text: 'Allocation: ' + convertToTitleCase(designInfo.allocation) });
        }
        if (designInfo.timePerspective && !types.includes('Time Perspective')) {
            categories.push({ text: 'Time Perspective: ' + convertToTitleCase(designInfo.timePerspective) });
        }
        if (designInfo.observationalModel && !types.includes('Observation Model')) {
            categories.push({ text: 'Observation Model: ' + convertToTitleCase(designInfo.observationalModel) });
        }
    }
    if (categories.length >= 1)
        result.category = categories;
    // ------- Category
    // Right now, the default value for a research study is "active". If CT.G
    // knows better, then allow it to override that.
    if (!result.status || result.status == 'active') {
        const overallStatus = (_f = protocolSection.statusModule) === null || _f === void 0 ? void 0 : _f.lastKnownStatus;
        if (overallStatus) {
            const status = convertClincalStudyStatusToFHIRStatus(overallStatus);
            if (typeof status !== 'undefined')
                result.status = status;
        }
    }
    if (!result.condition) {
        if ((_g = protocolSection.conditionsModule) === null || _g === void 0 ? void 0 : _g.conditions) {
            result.condition = convertArrayToCodeableConcept((_h = protocolSection.conditionsModule) === null || _h === void 0 ? void 0 : _h.conditions);
        }
    }
    if (!result.site) {
        const locations = (_j = protocolSection.contactsLocationsModule) === null || _j === void 0 ? void 0 : _j.locations;
        if (locations) {
            let index = 0;
            for (const location of locations) {
                if (typeof location === 'object' && location != null) {
                    const fhirLocation = { resourceType: 'Location', id: 'location-' + index++ };
                    if (location.status)
                        fhirLocation.status = convertRecruitmentStatusToLocationStatus(location.status);
                    if (location.facility)
                        fhirLocation.name = location.facility;
                    if (location.city && location.country) {
                        // Also add the address information
                        fhirLocation.address = { use: 'work', city: location.city, country: location.country };
                        if (location.state) {
                            fhirLocation.address.state = location.state;
                        }
                        if (location.zip) {
                            fhirLocation.address.postalCode = location.zip;
                        }
                    }
                    // If we have an exact GPS coordinate, add that, too
                    const geoPoint = location.geoPoint;
                    if (geoPoint && typeof geoPoint.lat === 'number' && typeof geoPoint.lon === 'number') {
                        fhirLocation.position = {
                            longitude: geoPoint.lon,
                            latitude: geoPoint.lat
                        };
                    }
                    if (location.contacts) {
                        for (const contact of location.contacts) {
                            if (contact.email) {
                                (0, research_study_1.addToContainer)(fhirLocation, 'telecom', {
                                    system: 'email',
                                    value: contact.email,
                                    use: 'work'
                                });
                            }
                            if (contact.phone) {
                                (0, research_study_1.addToContainer)(fhirLocation, 'telecom', {
                                    system: 'phone',
                                    value: contact.phone,
                                    use: 'work'
                                });
                            }
                        }
                    }
                    (0, research_study_1.addToContainer)(result, 'site', (0, research_study_1.addContainedResource)(result, fhirLocation));
                }
            }
        }
    }
    if (!result.arm) {
        const armGroups = (_k = protocolSection.armsInterventionsModule) === null || _k === void 0 ? void 0 : _k.armGroups;
        if (armGroups) {
            for (const studyArm of armGroups) {
                const label = studyArm.label;
                if (label) {
                    const arm = Object.assign(Object.assign({ name: label }, (studyArm.type && {
                        type: {
                            coding: [
                                {
                                    // It's unclear if there is any coding system for this, so for now, make it look like FHIR
                                    code: studyArm.type.replace(/_/g, '-').toLowerCase(),
                                    display: convertToTitleCase(studyArm.type)
                                }
                            ],
                            text: convertToTitleCase(studyArm.type)
                        }
                    })), (studyArm.description && { description: studyArm.description }));
                    (0, research_study_1.addToContainer)(result, 'arm', arm);
                }
            }
        }
    }
    if (!result.protocol) {
        const interventions = (_l = protocolSection.armsInterventionsModule) === null || _l === void 0 ? void 0 : _l.interventions;
        if (interventions) {
            let index = 0;
            for (const intervention of interventions) {
                if (intervention.armGroupLabels) {
                    for (const armGroupLabel of intervention.armGroupLabels) {
                        let plan = { resourceType: 'PlanDefinition', status: 'unknown', id: 'plan-' + index++ };
                        plan = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, plan), (intervention.description && { description: intervention.description })), (intervention.name && { title: intervention.name })), (intervention.otherNames &&
                            intervention.otherNames.length > 0 && { subtitle: intervention.otherNames[0] })), (intervention.type && { type: { text: convertToTitleCase(intervention.type) } })), { subjectCodeableConcept: { text: armGroupLabel } });
                        (0, research_study_1.addToContainer)(result, 'protocol', (0, research_study_1.addContainedResource)(result, plan));
                    }
                }
                else {
                    let plan = { resourceType: 'PlanDefinition', status: 'unknown', id: 'plan-' + index++ };
                    plan = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, plan), (intervention.description && { description: intervention.description })), (intervention.name && { title: intervention.name })), (intervention.otherNames &&
                        intervention.otherNames.length > 0 && { subtitle: intervention.otherNames[0] })), (intervention.type && { type: { text: convertToTitleCase(intervention.type) } }));
                    (0, research_study_1.addToContainer)(result, 'protocol', (0, research_study_1.addContainedResource)(result, plan));
                }
            }
        }
    }
    if (!result.contact) {
        const contacts = (_m = protocolSection.contactsLocationsModule) === null || _m === void 0 ? void 0 : _m.centralContacts;
        if (contacts) {
            for (const contact of contacts) {
                if (contact != undefined) {
                    const contactName = contact.name;
                    if (contactName) {
                        const fhirContact = { name: contactName };
                        if (contact.email) {
                            (0, research_study_1.addToContainer)(fhirContact, 'telecom', {
                                system: 'email',
                                value: contact.email,
                                use: 'work'
                            });
                        }
                        if (contact.phone) {
                            (0, research_study_1.addToContainer)(fhirContact, 'telecom', {
                                system: 'phone',
                                value: contact.phone,
                                use: 'work'
                            });
                        }
                        (0, research_study_1.addToContainer)(result, 'contact', fhirContact);
                    }
                }
            }
        }
    }
    if (!result.period) {
        const startDate = (_p = (_o = protocolSection.statusModule) === null || _o === void 0 ? void 0 : _o.startDateStruct) === null || _p === void 0 ? void 0 : _p.date;
        const completionDate = (_r = (_q = protocolSection.statusModule) === null || _q === void 0 ? void 0 : _q.completionDateStruct) === null || _r === void 0 ? void 0 : _r.date;
        if (startDate || completionDate) {
            // Set the period object as appropriate
            const period = Object.assign(Object.assign({}, (startDate && (0, fhir_type_guards_1.isFhirDate)(startDate) && { start: startDate })), (completionDate && (0, fhir_type_guards_1.isFhirDate)(completionDate) && { end: completionDate }));
            if (Object.keys(period).length != 0)
                result.period = period;
        }
    }
    return result;
}
exports.updateResearchStudyWithClinicalStudy = updateResearchStudyWithClinicalStudy;
